from PyQt6.QtWidgets import QLabel
from trilent.Utility import Reloader, qt_orientation, V, Misc
from trilent.Widgets import PositionTypes


class Text(Misc):
    def __init__(self,
                 parent,
                 # Special
                 text       : str       = None,
                 orientation: str       = None,
                 text_size  : str | int = None,
                 font       : str       = None,
                 width      : int | str = None,
                 height     : int | str = None,
                 auto_size  : bool      = None,  # Controls if you don't want to enter width and height
                 wrap       : bool      = None,

                 # Stylesheet
                 text_color      : str = None,
                 background_color: str = None):

        # Reloader setup
        def text_func(v):
            if self._reloader.cp['auto_size'] is True:
                self._widget.adjustSize()

            self._widget.setText(v)

        def orient_func(v): self._widget.setAlignment(v)

        self._reloader = Reloader(parent.get_dpi(),
                                  setup_properties  = {'wrap': wrap,                                  'background_color': background_color, 'text_color': text_color,   'auto_size': auto_size,   'text_size': text_size,     'font': font,         'orientation': orientation,   'text': text,            'parent': parent,   'width': width,          'height': height,},
                                  process_types     = {'wrap': 'super raw',                           'background_color': 'color',          'text_color': 'color',      'auto_size': 'super raw', 'text_size': 'px-value',    'font': 'raw',        'orientation': 'orientation', 'text': 'raw',           'parent': None,     'width': 'px-value;int', 'height': 'px-value;int'},
                                  default_values    = {'wrap': True,                                  'background_color': 'transparent',    'text_color': 'white',      'auto_size': False,       'text_size': '0.3 inch',    'font': 'Arial',      'orientation': 'top left',    'text': 'Trilent Text.', 'parent': None,     'width': '5 inch',       'height': '0.5 inch'},
                                  property_types    = {'wrap': 'special',                             'background_color': 'stylesheet',     'text_color': 'stylesheet', 'auto_size': 'access',    'text_size': 'stylesheet',  'font': 'stylesheet', 'orientation': 'special',     'text': 'special',       'parent': 'access', 'width': 'special',      'height': 'special'},
                                  special_functions = {'wrap': lambda v: self._widget.setWordWrap(v), 'orientation':       orient_func,     'text': text_func,          'width' : lambda v: self._widget.setGeometry(self.x, self.y, v, self.height), 'height': lambda v: self._widget.setGeometry(self.x, self.y, self.width, v)},
                                  base              = f"background-color: {V}background_color{V}; font-family: {V}font{V}; font-size: {V}text_size{V}; color: {V}text_color{V}")

        # Protected widget specific members
        # noinspection PyProtectedMember
        self._position_self = self._reloader.cp['parent']._position_children
        self._position_children = PositionTypes.NONE
        self._dpi = parent.get_dpi()

        # Qt setup
        # noinspection PyProtectedMember
        self._widget = QLabel(self._reloader.cp['text'], self._reloader.cp['parent']._get_holder())
        self._widget.setAlignment(qt_orientation(self._reloader.cp['orientation']))

        # Setup properties
        self._widget.setStyleSheet(self._reloader.reload())

        # Check if Box's parent is also a box
        # noinspection PyProtectedMember
        if self._reloader.cp['parent']._position_children == PositionTypes.BOX:
            # noinspection PyProtectedMember
            self._reloader.cp['parent']._widget_box_add(self)




from typing import Any


class PropertyManager:
    def __init__(self, kwargs):
        self.properties = {}

        for name, value in kwargs.items():
            self.properties[name] = value

    def get_property(self, name) -> Any:
        return self.properties[name]

    def get_properties(self, *args) -> tuple[Any, ...]:
        return tuple(self.get_property(name) for name in args)

    def set_property(self, name, value):
        assert name in self.properties, ValueError(f'Invalid property name: {name}, Available properties are: {list(self.properties.keys())}')
        self.properties[name] = value

    def set_properties(self, **kwargs):
        for name, value in kwargs.items():
            self.set_property(name, value)

    def new_property(self, name, initial_value):
        assert name not in self.properties, ValueError(f'Property name already exists: {name}')
        self.properties[name] = initial_value

    def new_properties(self, **kwargs):
        for name, value in kwargs.items():
            self.new_property(name, value)

    def __getitem__(self, key):
        return self.get_property(key)

    def __setitem__(self, key, value):
        self.set_property(key, value)



__all__ = ['Box', 'Button', 'Canvas', 'CheckBox', 'Entry', 'Slider', 'Switch', 'Text', 'Widget', 'Window', 'Animation']

from trilent.Widgets import *
from trilent.Animations import Animation
from trilent.Window import Window

_ = Box, Button, Canvas, CheckBox, Entry, Slider, Switch, Text, Widget, Window, Animation



import math


class EasingFunctions:
    @staticmethod
    def ease_in(t):
        return t * t * t

    @staticmethod
    def ease_out(t):
        t -= 1
        return t * t * t + 1

    @staticmethod
    def ease_in_out(t):
        t *= 2
        if t < 1:
            return 0.5 * t * t * t
        t -= 2
        return 0.5 * (t * t * t + 2)

    @staticmethod
    def back_out(t, s=1.70158):
        t -= 1
        return t * t * ((s + 1) * t + s) + 1

    @staticmethod
    def back_in(t, s=1.70158):
        return t * t * ((s + 1) * t - s)

    @staticmethod
    def back_in_out(t, s=1.70158):
        t *= 2
        if t < 1:
            return 0.5 * (t * t * ((s + 1) * t - s))
        t -= 2
        return 0.5 * (t * t * ((s + 1) * t + s) + 2)

    @staticmethod
    def sine_in(t):
        return 1 - math.cos((t * math.pi) / 2)

    @staticmethod
    def sine_out(t):
        return math.sin((t * math.pi) / 2)

    @staticmethod
    def sine_in_out(t):
        return -(math.cos(math.pi * t) - 1) / 2

    @staticmethod
    def quad_in(t):
        return t * t

    @staticmethod
    def quad_out(t):
        return t * (2 - t)

    @staticmethod
    def quad_in_out(t):
        return 2 * t * t if t < 0.5 else -1 + (4 - 2 * t) * t

    @staticmethod
    def cubic_in(t):
        return t * t * t

    @staticmethod
    def cubic_out(t):
        return (--t) * t * t + 1

    @staticmethod
    def cubic_in_out(t):
        return 4 * t * t * t if t < 0.5 else (t - 1) * (2 * t - 2) * (2 * t - 2) + 1

    @staticmethod
    def quart_in(t):
        return t * t * t * t

    @staticmethod
    def quart_out(t):
        return 1 - (--t) * t * t * t

    @staticmethod
    def quart_in_out(t):
        return 8 * t * t * t * t if t < 0.5 else 1 - 8 * (--t) * t * t * t

    @staticmethod
    def quint_in(t):
        return t * t * t * t * t

    @staticmethod
    def quint_out(t):
        return 1 + (--t) * t * t * t * t

    @staticmethod
    def quint_in_out(t):
        return 16 * t * t * t * t * t if t < 0.5 else 1 + 16 * (--t) * t * t * t * t

    @staticmethod
    def expo_in(t):
        return 0 if t == 0 else math.pow(2, 10 * (t - 1))

    @staticmethod
    def expo_out(t):
        return 1 if t == 1 else 1 - math.pow(2, -10 * t)

    @staticmethod
    def expo_in_out(t):
        if t == 0:
            return 0
        if t == 1:
            return 1
        return math.pow(2, 10 * (2 * t - 1)) / 2 if t < 0.5 else (2 - math.pow(2, -10 * (2 * t - 1))) / 2

    @staticmethod
    def circle_in(t):
        return 1 - math.sqrt(1 - t * t)

    @staticmethod
    def circle_out(t):
        return math.sqrt(1 - (--t) * t)

    @staticmethod
    def circle_in_out(t):
        return (1 - math.sqrt(1 - 4 * t * t)) / 2 if t < 0.5 else (math.sqrt(1 - (2 * t - 2) * (2 * t - 2)) + 1) / 2

    @staticmethod
    def linear(t):
        return t

    @staticmethod
    def linear_in(t):
        return t

    @staticmethod
    def linear_out(t):
        return t

    @staticmethod
    def linear_in_out(t):
        return t



from trilent.Widgets import Widget


class Spacer(Widget):
    def __init__(self,
                 parent,
                 # Special
                 width      : int | str = None,
                 height     : int | str = None):
        super().__init__(parent, width, height, widget_color='transparent')



from trilent.Utility.color_gradients import color_gradient, gradient_gradient
from trilent.Utility.other_gradients import text_gradient
from .helpers import better_range, break_px
from typing import Any


def generate_lists(name: str, type: str, start_value: Any, end_value: Any, mode: str = "") -> list:
    if type == 'px-value;int':
        # No need to break them up, Since it comes processed as ints.
        return better_range(start_value, end_value)
    elif type == 'px-value':
        # Break up the strings using break px,
        # Because the values often come processed as strings.
        return better_range(break_px(start_value), break_px(end_value))
    elif type == 'gradient color':
        # The more the steps/color_steps, The more the calculation times.
        return gradient_gradient(start_value, end_value, 200, 30)
    elif type == 'color':
        return color_gradient(start_value, end_value, 255)
    elif type == 'text':
        return text_gradient(start_value, end_value, mode)
    else:
        raise NotImplementedError(f'Animations for "{name}" isn\'t supported yet. Type: {type}')



from functools import lru_cache
import multiprocessing
from typing import List
from .color import get_as_qt


def split_rgb_format(rgb_string: str):
    color_channels = [int(channel) for channel in
                      rgb_string.removeprefix('rgb(').removeprefix('rgba(')[:-1].split(', ')]

    if len(color_channels) == 3:
        color_channels.append(255)

    return color_channels


def convert_hex_to_rgba(hex_color: str) -> list[int]:
    if hex_color.startswith('rgb(') or hex_color.startswith('rgba('):
        return split_rgb_format(hex_color)

    hex_color = f"{hex_color[1:]}FF"

    rgba = [int((hex_color + 'FF')[0:2], 16),
            int((hex_color + 'FF')[2:4], 16),
            int((hex_color + 'FF')[4:6], 16),
            int((hex_color + 'FF')[6:8], 16)]

    return rgba


def convert_rgba_to_hex(r, g, b, a=255) -> str: return f'#{r:02x}{g:02x}{b:02x}{a:02x}'


@lru_cache(30)
def color_gradient(color1: str | tuple[int, int, int], color2: str | tuple[int, int, int], steps: int = 10) -> list[
    str]:
    color1 = convert_hex_to_rgba(color1)
    color2 = convert_hex_to_rgba(color2)

    steps -= 1

    step_r = (color2[0] - color1[0]) / steps
    step_g = (color2[1] - color1[1]) / steps
    step_b = (color2[2] - color1[2]) / steps
    step_a = (color2[3] - color1[3]) / steps

    gradient = [
        convert_rgba_to_hex(int(color1[0] + step_r * i), int(color1[1] + step_g * i), int(color1[2] + step_b * i),
                            int(color1[3] + step_a * i)) for i in range(steps + 1)]

    return gradient


def multi_color_gradient(colors: list[str | tuple[int, int, int, int]], steps: int = 10) -> list[str]:
    gradient = []
    num_colors = len(colors)

    for i in range(num_colors - 1):
        color1 = convert_hex_to_rgba(colors[i])
        color2 = convert_hex_to_rgba(colors[i + 1])

        steps_between = steps - 1

        step_r = (color2[0] - color1[0]) / steps_between
        step_g = (color2[1] - color1[1]) / steps_between
        step_b = (color2[2] - color1[2]) / steps_between
        step_a = (color2[3] - color1[3]) / steps_between

        for j in range(steps_between):
            r = int(color1[0] + step_r * j)
            g = int(color1[1] + step_g * j)
            b = int(color1[2] + step_b * j)
            a = int(color1[3] + step_a * j)
            gradient.append(convert_rgba_to_hex(r, g, b, a))

        gradient.append(convert_rgba_to_hex(*color2))

    return gradient


def blend_color(color1: str, color2: str, step_ratio: float) -> str:
    c1 = convert_hex_to_rgba(color1)
    c2 = convert_hex_to_rgba(color2)
    blended_color = (int(c1[0] * (1 - step_ratio) + c2[0] * step_ratio),
                     int(c1[1] * (1 - step_ratio) + c2[1] * step_ratio),
                     int(c1[2] * (1 - step_ratio) + c2[2] * step_ratio),
                     int(c1[3] * (1 - step_ratio) + c2[3] * step_ratio))
    return convert_rgba_to_hex(blended_color[0], blended_color[1], blended_color[2], blended_color[3])


def blend_gradient_step(step: int, grad1: List[str], grad2: List[str], steps: int) -> str:
    step_ratio = step / steps
    blended_colors = [blend_color(color1, color2, step_ratio) for color1, color2 in zip(grad1, grad2)]
    return " -> ".join(blended_colors)


def blend_gradients(grad1: List[str], grad2: List[str], steps: int) -> List[str]:
    with multiprocessing.Pool() as pool:
        tasks = [(i, grad1, grad2, steps) for i in range(steps + 1)]
        blended_gradients = pool.starmap(blend_gradient_step, tasks)
    return blended_gradients


def gradient_gradient(gradient1: str, gradient2: str, steps: int, color_steps: int) -> list[str]:
    gradient1_colors = [get_as_qt(color) for color in gradient1.split(' -> ')]
    gradient2_colors = [get_as_qt(color) for color in gradient2.split(' -> ')]

    grad1 = multi_color_gradient(gradient1_colors, color_steps)
    grad2 = multi_color_gradient(gradient2_colors, color_steps)

    result = blend_gradients(grad1, grad2, steps)

    return result



from enum import Enum
from trilent.Animations import Animation
from PyQt6.QtWidgets import QGraphicsDropShadowEffect, QPushButton
from PyQt6.QtGui import QColor, QPainter
from PyQt6.QtCore import Qt, QBuffer, QIODevice
from PyQt6.QtSvg import QSvgGenerator
from .color import get_as_raw_qt
from typing import Any
from numpy import array


key_map = {
    Qt.Key.Key_Up: "up",
    Qt.Key.Key_Down: "down",
    Qt.Key.Key_Left: "left",
    Qt.Key.Key_Right: "right",
    Qt.Key.Key_Space: "space",
    Qt.Key.Key_Return: "enter",
    Qt.Key.Key_Backspace: "backspace",
    Qt.Key.Key_Tab: "tab",
    Qt.Key.Key_Escape: "escape",
    Qt.Key.Key_Delete: "delete",
    Qt.Key.Key_Home: "home",
    Qt.Key.Key_End: "end",
    Qt.Key.Key_PageUp: "page up",
    Qt.Key.Key_PageDown: "page down",
    Qt.Key.Key_CapsLock: "caps lock",
    Qt.Key.Key_Shift: "shift",
    Qt.Key.Key_Control: "control",
    Qt.Key.Key_Alt: "alt",
    Qt.Key.Key_Meta: "meta",
    Qt.Key.Key_Plus: "plus",
    Qt.Key.Key_Minus: "minus",
    Qt.Key.Key_Asterisk: "asterisk",
    Qt.Key.Key_Period: "period",
    Qt.Key.Key_Comma: "comma",
    Qt.Key.Key_Semicolon: "semicolon",
    Qt.Key.Key_Slash: "slash",
    Qt.Key.Key_Apostrophe: "apostrophe",
    Qt.Key.Key_QuoteDbl: "quote double",
}


def get_event_as_text(e):
    text = e.text()
    if text:
        return text
    else:
        return key_map.get(e.key(), 'unknown')


events_dictionary: dict[str, str] = {'wheel': 'wheelEvent',
                                     'drag': 'mouseMoveEvent',
                                     'hover': 'enterEvent',
                                     'unhover': 'leaveEvent',
                                     'clicked': 'mousePressEvent',
                                     'unclicked': 'mouseReleaseEvent',
                                     'key press': 'keyPressEvent',
                                     'key release': 'keyReleaseEvent'}


events_mappers: dict[str, Any] = {'wheel': lambda e, f: f(),
                                  'drag': lambda e, f: f(),
                                  'hover': lambda e, f: f(),
                                  'unhover': lambda e, f: f(),
                                  'clicked': lambda e, f: f(),
                                  'unclicked': lambda e, f: f(),
                                  'key press': lambda e, f: f(get_event_as_text(e)),
                                  'key release': lambda e, f: f(get_event_as_text(e))}


class PositionTypes(Enum):
    BOX: str = 'box'
    PLACE: str = 'place'
    NONE: str = 'None specified.'


class Misc:
    def __init__(self):
        self._widget = QPushButton()

    def place(self, x: int = 100, y: int = 100):
        assert self._reloader.cp['parent']._position_children != PositionTypes.BOX, \
            TypeError("You cant place a widget whose parent is a box. Instead, Its position is automatically handled.")

        geometry = self._widget.geometry()
        self._widget.setGeometry(x, y, geometry.width(), geometry.height())
        self._widget.show()

    def force_place(self, x: int = 100, y: int = 100):
        geometry = self._widget.geometry()
        self._widget.setGeometry(x, y, geometry.width(), geometry.height())
        self._widget.show()

    def set_tooltip(self, tooltip): self._widget.setToolTip(tooltip)
    def get_tooltip(self): self._widget.toolTip()

    def set_position(self, x, y): self._widget.setGeometry(x, y, self.width, self.height)
    def set_size(self, width, height): self._widget.setGeometry(self.x, self.y, width, height)

    def get_position(self):
        geometry = self._widget.geometry()
        return geometry.x(), geometry.y()

    def get_size(self):
        geometry = self._widget.geometry()
        return geometry.width(), geometry.height()

    def change(self, **kwargs):
        for k, v in kwargs.items():
            self._reloader.set(k, v)

        self._widget.setStyleSheet(self._reloader.reload())

    def set(self, name, value):
        self._reloader.set(name, value)

        ss = self._reloader.reload()

        if self._reloader.initial_parameters['property_types'][name] == 'stylesheet':
            self._widget.setStyleSheet(ss)

    def get(self, property_name):
        return self._reloader.cp[property_name]

    @property
    def width(self): return self._widget.width()
    @property
    def height(self): return self._widget.height()
    @property
    def x(self): return self._widget.x()
    @property
    def y(self): return self._widget.y()

    def show(self): self._widget.show()
    def hide(self): self._widget.hide()

    def is_hidden(self): return self._widget.isHidden()
    def is_shown(self): return not self._widget.isHidden()

    def get_top_parent(self): return self.get('parent').get_top_parent()

    def connect(self, name, func):
        caller = lambda e: events_mappers[name](e, func)
        setattr(self._widget, events_dictionary[name], caller)

    def animate(self, time=0.3, curve='ease in out', animation_finished=lambda: ..., mode='', **kwargs):
        for i, (k, v) in enumerate(kwargs.items()):
            Animation(self, k, v, time=time, curve=curve, mode=mode, animation_finished=animation_finished if i == 0 else lambda: ...).start()

    def effect(self, name, **kwargs):
        if name == "shadow":
            blur, x,  y, color = kwargs['blur'], kwargs['x'], kwargs['y'], kwargs['color']
            effect = QGraphicsDropShadowEffect()
            effect.setXOffset(x)
            effect.setYOffset(y)
            effect.setBlurRadius(blur)
            effect.setColor(QColor(*get_as_raw_qt(color)))

            self._widget.setGraphicsEffect(effect)

    def remove_effects(self):
        self._widget.setGraphicsEffect(None)

    def enable(self): self._widget.setDisabled(False)
    def disable(self): self._widget.setDisabled(True)

    def add_update_function(self, f):
        self.get_top_parent().add_update_function(f)

    def get_pixel_data(self):
        pixmap = self._widget.grab()

        image = pixmap.toImage()

        width = image.width()
        height = image.height()

        ptr = image.bits()
        ptr.setsize(image.sizeInBytes())
        arr = array(ptr).reshape((height, width, 4))

        return arr

    def get_as_svg(self):
        buffer = QBuffer()
        buffer.open(QIODevice.OpenModeFlag.ReadWrite)

        svg_generator = QSvgGenerator()
        svg_generator.setOutputDevice(buffer)
        svg_generator.setSize(self._widget.size())
        svg_generator.setViewBox(self._widget.rect())
        svg_generator.setTitle("Widget as SVG")
        svg_generator.setDescription("Trilent SVG")

        painter = QPainter(svg_generator)
        self._widget.render(painter)
        painter.end()

        buffer.seek(0)
        return buffer.readData(buffer.size()).decode()

    def delete(self):
        self._widget.setParent(None)



from PyQt6.QtWidgets import QFrame
from trilent.Utility import Reloader, V, Misc
from trilent.Widgets import PositionTypes
from trilent.FlexComputer import box


class Box(Misc):
    def __init__(self,
                 parent,
                 # Utility
                 alignment: str = None,
                 side_alignment: str = None,
                 width: int | str = None,
                 height: int | str = None,

                 # Styling
                 # Color
                 box_color: str = None,
                 # Fancy
                 corner_roundness: int | str = None,

                 # Flex Properties
                 wrap: bool = None,
                 gap: int | str = None,
                 vertical_gap: int | str = None):

        def width_func(v):
            self._widget.setGeometry(self.x, self.y, v, self.height)

        def height_func(v):
            self._widget.setGeometry(self.x, self.y, self.width, v)

        # Reloader setup
        self._reloader = Reloader(parent.get_dpi(),
                                  setup_properties  = {'parent': parent,   'width': width,          'height': height,         'box_color': box_color,     'corner_roundness': corner_roundness, 'alignment': alignment, 'side_alignment': side_alignment, 'wrap': wrap,     'gap': gap,            'vertical_gap': vertical_gap},
                                  process_types     = {'parent': None,     'width': 'px-value;int', 'height': 'px-value;int', 'box_color': 'color',       'corner_roundness': 'px-value',       'alignment': None,      'side_alignment': None,           'wrap': None,     'gap': 'px-value;int', 'vertical_gap': 'px-value;int'},
                                  default_values    = {'parent': None,     'width': '3 inch',       'height': '2 inch',       'box_color': 'transparent', 'corner_roundness': '0.03 inch',      'alignment': 'start',   'side_alignment': 'start',        'wrap': True,     'gap': '1 px',         'vertical_gap': '1 px'},
                                  property_types    = {'parent': 'access', 'width': 'special',      'height': 'special',      'box_color': 'stylesheet',  'corner_roundness': 'stylesheet',     'alignment': 'access',  'side_alignment': 'access',       'wrap': 'access', 'gap': 'access',       'vertical_gap': 'access'},
                                  special_functions = {'width': width_func, 'height': height_func},
                                  base              = f"background-color: {V}box_color{V}; border-radius: {V}corner_roundness{V}")

        # Protected widget specific members
        # noinspection PyProtectedMember
        self._position_self = self._reloader.cp['parent']._position_children
        self._position_children = PositionTypes.BOX
        self._dpi = parent.get_dpi()
        self._widgets = []
        self._children = []

        # Qt setup
        # noinspection PyProtectedMember
        self._widget = QFrame(self._reloader.cp['parent']._get_holder())
        self._reloader.cp['parent'].add_update_function(self._update)

        # Setup properties
        self._widget.setStyleSheet(self._reloader.reload())

        # Check if Box's parent is also a box
        # noinspection PyProtectedMember
        if self._reloader.cp['parent']._position_children == PositionTypes.BOX:
            # noinspection PyProtectedMember
            self._reloader.cp['parent']._widget_box_add(self)

        self._update()

    def _get_holder(self): return self._widget

    def _widget_box_add(self, trilent_widget):
        self._children.append(trilent_widget)
        trilent_widget.force_place(0, 0)

    def _update(self):
        widths = tuple(child.width for child in self._children)
        heights = tuple(child.height for child in self._children)

        if len(widths) == 0 or len(heights) == 0:
            # Return early so the entire system doesn't crash
            # The current flexbox calculations assume that widths and heights aren't empty.
            return

        flex_args = (self._reloader.process(k, self._reloader.cp[k]) for k in ['alignment', 'side_alignment', 'wrap', 'gap', 'vertical_gap'])

        main_axis = box(widths, heights, self.width, self.height, *flex_args)

        for i, child in enumerate(self._children):
            child.set_position(main_axis[i][0], main_axis[i][1])

    def get_dpi(self): return self._dpi
    def add_update_function(self, f):
        self._reloader.cp['parent'].add_update_function(f)



from .start_alignment import calculate_start_positions
from .center_alignment import calculate_center_positions
from .end_alignment import calculate_end_positions


def box(widths: tuple[int, ...], heights: tuple[int, ...], box_width: int = 500, box_height: int = 500, alignment: str = 'start', side_alignment: str = 'start', wrap: bool = False, gap: int = 5, vertical_gap: int = 5):
    arguments = (widths, heights, (box_width, box_height), wrap, gap, vertical_gap, side_alignment)

    # Different functions for different scenarios because of speed.
    if alignment == 'start':
        return calculate_start_positions(*arguments)
    elif alignment == 'center':
        return calculate_center_positions(*arguments)
    elif alignment == 'end':
        return calculate_end_positions(*arguments)



CACHING_LIMIT = 10
FUNCTIONS = 9


def fallback(positions, heights, vertical_gap, side_alignment='start', flex_box_height=0) -> list[list[int]]:
    y = 0 if side_alignment in ['start', 'center'] else flex_box_height - heights[0]

    for i, pos in enumerate(positions):
        positions[i] = [pos[0], y]

        if side_alignment != 'end':
            y += heights[i] + vertical_gap
        else:
            y -= vertical_gap + heights[i]

    return positions


def center_y_positions(positions: list[list[int]], flex_box_size: tuple[int, int], total_height: int) -> list[list[int]]:
    adder = (flex_box_size[1] // 2) - (total_height // 2)
    return [[pos[0], pos[1] + adder] for pos in positions]




from .general import fallback, CACHING_LIMIT, FUNCTIONS, center_y_positions
from functools import lru_cache


@lru_cache(CACHING_LIMIT // FUNCTIONS)
def calculate_start_positions(widths: tuple[int, ...], heights: tuple[int, ...], flex_box_size: tuple[int | str, int | str], wrap: bool, gap: int, vertical_gap: int, side_alignment: str) -> list[list[int]]:
    if side_alignment == 'end':
        # Start y from the height of the container
        x, y = 0, flex_box_size[1]
    else:
        x, y = 0, 0

    positions: list[list[int]] = []
    column_heights: list[int] = []
    wraps = 1
    # Side alignment center
    total_height = 0

    for i, width in enumerate(widths):
        if wrap and x + width > flex_box_size[0] and i != 0:
            x = 0
            if side_alignment == 'end':
                y -= max(column_heights) + vertical_gap  # Move y downwards for wrapping
            else:
                y += max(column_heights) + vertical_gap
            total_height += max(column_heights) + vertical_gap

            wraps += 1
            column_heights = []

        if side_alignment == 'end':
            positions.append([x, y - heights[i]])
        else:
            positions.append([x, y])
        column_heights.append(heights[i])
        x += width + gap

    total_height += max(column_heights) + vertical_gap if column_heights else 0

    if wraps >= len(widths): positions = fallback(positions, heights, vertical_gap, side_alignment, flex_box_size[1])
    if side_alignment == 'center': positions = center_y_positions(positions, flex_box_size, total_height)

    return positions




from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    readme = "\n" + fh.read()

VERSION = '2.3.4.8'  # Increment on next change
DESCRIPTION = 'Python GUI Framework.'

setup(
    name="trilent",
    version=VERSION,
    author="Schkimansky",
    author_email="<ahmadchawla1432@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=readme,
    packages=find_packages(),
    install_requires=['PyQt6', 'conversa', 'numpy'],
    keywords=['python', 'gui', 'ui', 'framework', 'app', 'application', 'ui building', 'app development'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)



from PyQt6.QtWidgets import QLineEdit
from trilent.Utility import Reloader, V, Misc
from trilent.Widgets import PositionTypes


class Entry(Misc):
    def __init__(self,
                 parent,
                 # Special
                 text       : str       = None,
                 text_size  : str | int = None,
                 font       : str       = None,
                 width      : int | str = None,
                 height     : int | str = None,
                 command    : lambda: ... = None,

                 # Stylesheet
                 text_color      : str = None,
                 background_color: str = None,
                 corner_roundness: str | int = None,
                 hover_color     : str = None,
                 clicked_color   : str = None,

                 border_width    : str | int = None,
                 border_color    : str       = None):

        # Reloader setup
        def text_func(v):
            self._widget.setText(v)

        base = f"""QLineEdit {{ background-color: {V}background_color{V}; border: {V}border_width{V} solid {V}border_color{V}; font-family: {V}font{V}; font-size: {V}text_size{V}; color: {V}text_color{V}; border-radius: {V}corner_roundness{V}; }}
QLineEdit:hover {{ background-color: {V}hover_color{V}; }}
QLineEdit:focus {{ background-color: {V}clicked_color{V}; }}"""

        self._reloader = Reloader(parent.get_dpi(),
                                  setup_properties  = {'border_width': border_width,  'border_color': border_color,  'hover_color': hover_color,   'clicked_color': clicked_color,    'command': command,     'corner_roundness': corner_roundness, 'background_color': background_color, 'text_color': text_color,   'text_size': text_size,     'font': font,         'text': text,              'parent': parent,   'width': width,          'height': height,},
                                  process_types     = {'border_width': 'px-value',    'border_color': 'color',       'hover_color': 'color',       'clicked_color': 'color',          'command': 'super raw', 'corner_roundness': 'px-value',       'background_color': 'color',          'text_color': 'color',      'text_size': 'px-value',    'font': 'raw',        'text': 'raw',             'parent': None,     'width': 'px-value;int', 'height': 'px-value;int'},
                                  default_values    = {'border_width': '0.0105 inch', 'border_color': 'litecrimson', 'hover_color': 'litecrimson', 'clicked_color': 'delightcrimson', 'command': lambda: ..., 'corner_roundness': '0.04 inch',      'background_color': 'crimson',        'text_color': 'white',      'text_size': '0.20 inch',   'font': 'Verdana',    'text': 'Trilent Entry.',  'parent': None,     'width': '1.75 inch',    'height': '0.45 inch'},
                                  property_types    = {'border_width': 'stylesheet',  'border_color': 'stylesheet',  'hover_color': 'stylesheet',  'clicked_color': 'stylesheet',     'command': 'access',    'corner_roundness': 'stylesheet',     'background_color': 'stylesheet',     'text_color': 'stylesheet', 'text_size': 'stylesheet',  'font': 'stylesheet', 'text': 'special',         'parent': 'access', 'width': 'special',      'height': 'special'},
                                  special_functions = {'text': text_func,  'width': lambda v: self._widget.setGeometry(self.x, self.y, v, self.height), 'height': lambda v: self._widget.setGeometry(self.x, self.y, self.width, v)},
                                  base              = base)

        # Protected widget specific members
        # noinspection PyProtectedMember
        self._position_self = self._reloader.cp['parent']._position_children
        self._position_children = PositionTypes.NONE
        self._dpi = parent.get_dpi()

        # Qt setup
        # noinspection PyProtectedMember
        self._widget = QLineEdit(self._reloader.cp['parent']._get_holder())
        self._widget.setText(self._reloader.cp['text'])

        # Setup properties
        self._widget.setStyleSheet(self._reloader.reload())
        self._widget.editingFinished.connect(lambda: self._reloader.cp['command']())

        # Check if Box's parent is also a box
        # noinspection PyProtectedMember
        if self._reloader.cp['parent']._position_children == PositionTypes.BOX:
            # noinspection PyProtectedMember
            self._reloader.cp['parent']._widget_box_add(self)

    def get(self, property_name):
        if property_name == 'text':
            return self._widget.text()
        return self._reloader.cp[property_name]




def better_range(start_value, end_value):
    if start_value < end_value:
        return list(range(start_value, end_value + 1))
    elif start_value == end_value:
        return [end_value]
    elif start_value > end_value:
        return list(range(start_value, end_value, -1))


def break_px(string: str):
    """ Transforms '10px' to 10 """
    return int(float(string.removesuffix('px')))
